/****************************************************************************
 *                                                          __              *
 *                ____ ___  ____ __  ______ ___  ____  ____/ /              *
 *               / __ `__ \/ __ `/ |/ / __ `__ \/ __ \/ __  /               *
 *              / / / / / / /_/ />  </ / / / / / /_/ / /_/ /                *
 *             /_/ /_/ /_/\__,_/_/|_/_/ /_/ /_/\____/\__,_/                 *
 *                                                                          *
 *                             GBA Sound Test                               *
 *                                                                          *
 *         Copyright (c) 2008, Mukunda Johnson (mukunda@maxmod.org)         *
 *                                                                          *
 * Permission to use, copy, modify, and/or distribute this software for any *
 * purpose with or without fee is hereby granted, provided that the above   *
 * copyright notice and this permission notice appear in all copies.        *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES *
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF         *
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR  *
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   *
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    *
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF  *
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           *
 ****************************************************************************/

#include "mp_mas_structs.inc"
#include "mp_format_mas.inc"

/****************************************************************************
 *
 * DEFINITIONS
 *
 ****************************************************************************/

// Screen Layout

/******************************
 *                  <------------ DMA0: bg gradient
 *       Centered Title <---  *
 *                         |  *
 *     +----------------+  |  *
 *     |    00:00:00 <------  *
 *     |          <--------|----- BG2: window display
 *     |  Music:    00<----+----- BG0+BG1: text + shadow
 *.....|                |#.|..*
 *.....|  Pattern:  00  |#.|..*
 *.....|                |#<|----- BG3: window shadow
 *.....|  Tempo:  100%  |#.|..*
 *.....|                |#.|..*
 *.....|  Pitch:  100%  |#.|..*
 *:::::|                |#:|::*
 *:::::+----------------+#:|::*
 *::::::##################:|::*
 *:::::::::::::::::::::::::|::*
 *#####Press Start to Exit<-##*
 *############################*
 ******************************/

// Palette layout

/*********************************************************
 | background, 10 colors | text color | shadow | ... |
 ---------------------------------------------------------
 | x | x | window face | window border | ... |
 *******************************************/

.EQU	DISPCNT,	0x000
.EQU	DISPSTAT,	0x004
.EQU	IE,		0x200
.EQU	IME,		0x208
.EQU	WAITCNT,	0x204
.EQU	BG0CNT,		0x008
.EQU	BG0HOFS,	0x010
.EQU	BG0VOFS,	0x012
.EQU	BLDCNT,		0x050
.EQU	COLEY,		0x054
.equ	PALETTE,	0x5000000

.equ	DMA0SAD,	0x40000B0
.equ	KEYINPUT,	0x4000130

.equ	BG0MAP,		0x6001800
.equ	BG1MAP,		0x6002000
.equ	BG2MAP,		0x6002800
.equ	BG3MAP,		0x6003000

.equ	CpuSet,		0x0B
.equ	CpuSetArm,		0x0B0000
.equ	BitUnPack,	0x10
.equ	VBlankIntrWait,	0x05
.equ	VBlankIntrWaitArm,	0x050000
.equ	Div,		0x06

/*.equ	SB_ADDR,	0x3400
.equ	SB_ADDR_RAM,	0x2000000+SB_ADDR
.equ	SB_ADDR_CART,	0x8000000+SB_ADDR
.EQU	MAX_RAMDATA,	(262144 - SB_ADDR)
*/

.equ	ARGUMENTS_OFFSET, 0x0203fbf8
.equ	ARGUMENTS_OFFSET_32BYTE_ALIGNED, (0x0203fbf8/32*32)
.equ	POGOSHELL_MAGIC_OFFSET, (ARGUMENTS_OFFSET + 8)
.equ	POGOSHELL_MAGIC, 0xFAB0BABE
.equ	FILE_OFFSET, (ARGUMENTS_OFFSET + 4)
.equ	FILE_ARGUMENTS_OFFSET, (ARGUMENTS_OFFSET + 16) 
.equ	KEYS_ALL,	0x3ff

.equ	TEXT_COLOR,	0x7FFF0000
.equ	WINDOW_COLOR,	0x7FFF5D60


.equ	KEY_A,		1
.equ	KEY_B,		2
.equ	KEY_SELECT,	3
.equ	KEY_START,	4
.equ	KEY_RIGHT,	5
.equ	KEY_LEFT,	6
.equ	KEY_UP,		7
.equ	KEY_DOWN,	8
.equ	KEY_R,		9
.equ	KEY_L,		10

.equ	BUTT_A,		0
.equ	BUTT_B,		1
.equ	BUTT_LEFT,	2
.equ	BUTT_RIGHT,	3

.equ	PRESS_TIMEOUT,	7
.equ	SCROLL_RATE,	4

.equ	STATE_PLAY, 0
.equ	STATE_PAUSE, 1
.equ 	STATE_STOP, 2

/****************************************************************************
 *
 * MEMORY
 *
 ****************************************************************************/

.struct 0
v_keys_prev:	.space 2
v_music:	.space 2
v_pattern:	.space 2
v_tempo:	.space 2
v_pitch:	.space 2
v_ntracks:	.space 2
v_npatterns:	.space 2
v_uflags:	.space 1
v_playstate:	.space 1

v_sel:		.space 1
v_ptime:	.space 1
v_press:	.space 1
v_butt:		.space 1
v_cycles:	.space 4
v_time:		.space 4
v_size:

.equ	UF_MUSIC, 1
.equ	UF_PATTERN, 2
.equ	UF_TEMPO, 4
.equ	UF_PITCH, 8
.equ	UF_TIME, 16

	.BSS
	.ALIGN

mixing_buffer:
	.space	1408
wave_buffer:
	.space	1408

mod_ch:
	.space	40*32
act_ch:
	.space	28*32
mix_ch:
	.space	24*32

bg_color_data:
	.space 160*2

variables:
	.space v_size

/****************************************************************************
 * 
 * CARTRIDGE HEADER
 *
 ****************************************************************************/

	.SECTION ".header", "ax"
	.ARM

	.GLOBAL     _start
_start:
	b       rom_start	@ start vector (cart)
        .fill   156,1,0		@ nintendo logo
	.string	"MAXMOD DEMO"	@ game title
	.byte	0,0,0,0		@ game code
	.byte	0x30,0x31	@ maker code
	.byte	0x96		@ fixed value
	.byte	0,0		@ main unit code, device type
	.byte	0,0,0,0,0,0,0	@ unsed data
	.byte	1		@ software version
	.byte	0		@ complement check
	.byte	0,0		@ reserved
.arm
        b       unpack_program	@ start vector (multiboot)
	.byte	0,0		@ boot mode / slave_id

/****************************************************************************
 *
 * Setup
 *
 ****************************************************************************/

	.ALIGN

//----------------------------------------------------
// CART entry: copy to RAM
//----------------------------------------------------
rom_start:
//----------------------------------------------------
	adr r0, unpack_program
	tst r0, #0x8000000
	beq unpack_program

	mov	r0, #0x8000000
	mov	r1, #0x2000000
	ldr r2, =ARGUMENTS_OFFSET_32BYTE_ALIGNED
.copy2ram:
	ldmia	r0!, {r3-r10}
	stmia	r1!, {r3-r10}
	cmp r1,r2
	blt	.copy2ram
	
@---------------------------------------
unpack_program:
@---------------------------------------
	mov	r0, #0xFE		@ clear registers/ram
	swi	1<<16			@ SWI 1, REGISTER RAM CLEAR	
	
	ldr	r0,=__text_lma
	ldr	r1,=__text_start
	swi	0x12<<16		@ lz77 unpack
	
	ldr	pc,=__PROGRAM_START
.pool
	
/****************************************************************************
 *
 * Begin Compressed Program
 *
 ****************************************************************************/

	.TEXT
	.ARM
	.ALIGN

//---------------------------------------------------------------------------
__PROGRAM_START:
//---------------------------------------------------------------------------
	
	mov	r4, #0x4000000		@ registers
	ldr	r0,=intr_main		@ set user IRQ vector
	str	r0, [r4, #-4]		
	mrs	r0, cpsr		@ enable interrupts (CPU)
	bic	r0, r0, #0x80		
	msr	cpsr, r0		
	mov	r0, #0			@ setup communications for BURSTBOOT BACKDOOR
	str	r0, [r4, #0x134]	@ RCNT (normal mode)
	ldr	r0,=0x5080
	str	r0, [r4, #0x128]	@ SIOCNT (normal mode)
	ldr	r0, msg_boot		@ burstboot signal
	str	r0, [r4, #0x120]	@ SIODATA32
	mov	r0, #0x0008
	str	r0, [r4, #0x4]
	mov	r0, #1			@ IME=true
	str	r0, [r4, #0x208]

	mov	r0, #0x81		@ IE=vblank|comms
	str	r0, [r4, #0x200]

	mov	r0, #0x8
	strh	r0, [r4, #DISPSTAT]	@ enable vblank interrupt

	add	r0, pc, #1		@ switch to THUMB
	bx	r0

.THUMB

/****************************************************************************
 * AgbMain()
 *
 * Program entry point
 ****************************************************************************/
AgbMain:
	ldr	r0,=0x4000000+WAITCNT	// setup default ROM timing
	ldr	r1,=0x4014		// 3,1 cycles, enable prefetch
	strh	r1, [r0]		//

	ldr r0,=POGOSHELL_MAGIC_OFFSET
	ldr r0,[r0]
	ldr r1,=POGOSHELL_MAGIC
	cmp r0,r1
	beq 1f
	ldr r0,=reset
	bx r0

1:
	ldr r1,=FILE_OFFSET
	ldr r1,[r1]
	ldr	r0,=sndsystem		//
	str	r1, [r0, #32]		//
	push	{r1}
	bl	mmInit			//

//----------------------------------------
// setup some stuff
//----------------------------------------

	ldr	r7,=variables
	

	pop	{r5}			// r0 = neffects (total)

	ldrh	r1, [r5, #2]		// copy ntracks
	strh	r1, [r7, #v_ntracks]	// 
	
//----------------------------------------
// setup palette
//----------------------------------------

	ldr	r1,=PALETTE		// set window colors//
	ldr	r2,=WINDOW_COLOR	//
	str	r2, [r1, #2*2]		//
	ldr	r0,=TEXT_COLOR		// set text colors
	str	r0, [r1, #0*2]		//

//----------------------------------------
// generate bg gradient
//----------------------------------------

	ldr	r0,=31<<10
	ldr	r1,=1<<10
	ldr	r2,=bg_color_data
	mov	r3, #8
	mov	r4, #20
	
1:	strh	r0, [r2]
	add	r2, #2
	sub	r3, #1
	bne	1b
	sub	r0, r1
	mov	r3, #8
	sub	r4, #1
	bne	1b	

//----------------------------------------
// decompress font
//----------------------------------------
	
	ldr	r0,=fontTiles
	ldr	r1,=0x6000400
	ldr	r2,=unpack_tiles
	swi	BitUnPack
	
//----------------------------------------

	ldr	r0,=windowTiles
	ldr	r1,=0x6001000
	mov	r2, #288/2
	lsl	r2, #1
	swi	CpuSet

//----------------------------------------
// setup controls
//----------------------------------------

	ldr	r0,=bg_controls
	ldr	r1,=0x4000000+BG0CNT
	mov	r2, #12
	swi	CpuSet

//----------------------------------------
// draw window
//----------------------------------------

	ldr	r0,=BG2MAP
	mov	r1, #128

//---------------------------
// top
//---------------------------
	strh	r1, [r0]		// top left
	add	r0, #2
	
	add	r1, #1
	mov	r2, #20
	
1:	strh	r1, [r0]		// top
	add	r0, #2			//
	sub	r2, #1			//
	bne	1b			//

	ldr	r1,=128 + (1<<10)	// top right
	strh	r1, [r0]

//---------------------------
// middle
//---------------------------

	mov	r2, #12

2:	mov	r1, #128+2
	add	r0, #22
	strh	r1, [r0]		// left
	add	r0, #2
	add	r1, #1
	
	mov	r3, #20
1:	strh	r1, [r0]		// middle
	add	r0, #2
	sub	r3, #1
	bne	1b

	ldr	r1,=128+2 + (1<<10)
	strh	r1, [r0]		// right

	sub	r2, #1
	bne	2b

//---------------------------
// bottom
//---------------------------

	ldr	r1,=128 + (1<<11)
	add	r0, #22
	strh	r1, [r0]
	add	r1, #1
	add	r0, #2
	
	mov	r2, #20
1:	strh	r1, [r0]
	add	r0, #2
	sub	r2, #1
	bne	1b

	ldr	r1,=128 + (1<<10) + (1<<11)
	strh	r1, [r0]
	
//----------------------------------------
// copy to shadow
//----------------------------------------

	ldr	r1,=BG2MAP
	ldr	r0,=BG3MAP
	mov	r2, #1
	lsl	r2, #9
	ldr	r4,=0x10001000

1:	ldmia	r1!, {r3}
	add	r3, r4
	stmia	r0!, {r3}
	sub	r2, #1
	bne	1b

//----------------------------------------
// setup blending
//----------------------------------------
// blend bg1+bg3
// 50%,50%

	ldr	r0,=0x4000000+BLDCNT
	ldr	r1,=0x0808354A
	str	r1, [r0]

//----------------------------------------
// setup sprites
//----------------------------------------

	ldr	r0,=fingeTiles
	ldr	r1,=0x6010000
	mov	r2, #32/2
	swi	CpuSet
	
	ldr	r0,=fingePal
	ldr	r1,=0x5000200
	mov	r2, #16/2
	swi	CpuSet

	ldr	r0,=0x7000000		// disable all sprites
	mov	r1, #1			//
	lsl	r1, #9			//
	mov	r2, #128		//
					//
1:	stmia	r0!, {r1,r2}		//
	sub	r2, #1			//
	bne	1b			//

//----------------------------------------
// initialize information
//----------------------------------------


	ldr	r7,=variables

	mov	r0, #1
	lsl	r0, #10
	strh	r0, [r7, #v_tempo]
	strh	r0, [r7, #v_pitch]

	ldr r2, =FILE_ARGUMENTS_OFFSET
// skip command
.not_zero_yet:
	ldrb r0,[r2]
	add r2,r2,#1
	cmp r0,#0
	bne .not_zero_yet

// jump to end of filename
.not_zero_yet2:
	ldrb r0,[r2]
	add r2,r2,#1
	cmp r0,#0
	bne .not_zero_yet

	mov r3,#0
// and backtrack to last slash
.not_slash_yet:
	add r3,r3,#1
	sub r2,r2,#1
	ldrb r0,[r2]
	cmp r0,#0x2f
	bne .not_slash_yet
// then skip it
	add r2,r2,#1
	sub r3,r3,#1
	lsr r3,r3,#1

	mov r0,#16
	sub r0,r0,r3
	mov r1, #1
	bl textString

	mov r0, #11
	mov r1, #4
	adr r2, str_TIME
	bl textString

	mov r0, #6
	mov r1, #18
	adr r2, str_EXIT
	bl textString

	mov	r0, #7
	mov	r1, #7
	adr	r2, str_MUSIC
	bl	textString

	mov	r0, #7
	mov	r1, #9
	adr	r2, str_PATTERN
	bl	textString

	mov	r0, #7
	mov	r1, #11
	adr	r2, str_TEMPO
	bl	textString

	mov	r0, #7
	mov	r1, #13
	adr	r2, str_PITCH
	bl	textString
	
	mov	r1, #0xFF
	strb	r1, [r7, #v_uflags]

	bl	updateInfo

	bl	textUpdateShadow

	
	bl	updateFinge
	
	b	.enable_display
.align
str_EXIT:
	.string "Press Start to Exit"
.align
str_TIME:
	.string "--:--:--"
.align
str_MUSIC:
	.string "Music:        --"
.align
str_PATTERN:
	.string "Pattern:  --/ --"
.align
str_TEMPO:
	.string "Tempo:      100%"
.align
str_PITCH:
	.string "Pitch:      100%"

.align 2

//----------------------------------------
// enable display
//----------------------------------------
.enable_display:
	swi	VBlankIntrWait

	ldr	r0,=0x4000000+DISPCNT
	ldr	r1,=(1<<8) | (1<<9) | (1<<10) | (1<<11) | (1<<12)
	strh	r1, [r0]

	mov r0, #0
	strh r0, [r7, #v_pattern]
	bl music_start

//----------------------------------------
// main loop
//----------------------------------------

main_loop:

	swi	VBlankIntrWait
	
//----------------------------------------
// key input
//----------------------------------------
	
	ldr	r0,=KEYINPUT		// read keys [inversed bits]
	ldrh	r0, [r0]		//
	mvn	r5, r0			// 

	// keydown = 0->1 ((p AND k) XOR k)
	// keyheld = 1->1 (p AND k)
	// keyup   = 1->0 (k AND p) XOR p)

	// A  = 0
	// B   = 1
	// Left = 2
	// Right = 3

	ldrh	r4, [r7, #v_keys_prev]
	strh	r5, [r7, #v_keys_prev]
	
	mov	r6, r4			// r6 = keysdown
	and	r6, r5			//
	eor	r6, r5			//

	lsr	r0, r6, #KEY_UP		// up: move cursor up
	bcc	1f			//
	bl	selUp			//
1:
	lsr	r0, r6, #KEY_DOWN	// down: move cursor down
	bcc	1f			//
	bl	selDown			//
1:

	mov	r0, r6			// do keysdown actions
	ldr	r1,=selPress		//
	bl	keyActions		//

	mov	r0, r4			// do keysheld actions
	and	r0, r5			// (maybe change to just r0=r5 ?)
	ldr	r1,=selHold		//
	bl	keyActions		//
	
	mov	r0, r5			// do keysup actions
	and	r0, r4			//
	eor	r0, r4			//
	ldr	r1,=selRelease		//
	bl	keyActions		//

	bl	updateFinge

	ldrb	r0, [r7, #v_playstate] // test&update time
	cmp r0, #STATE_PLAY
	bne 2f

	ldr r0, =mmLayerMain
	ldrb	r0, [r0, #MPL_POSITION]
	ldrh	r1, [r7, #v_pattern]
	cmp r0, r1
	beq 1f

	strb	r0, [r7, #v_pattern]
	ldrb	r1, [r7, #v_uflags]
	mov		r2, #UF_PATTERN
	orr		r1, r2
	strb	r1, [r7, #v_uflags]
1:

	ldr	r0, [r7, #v_cycles]
	ldr r1, =280896			// 280896 CPU cycles in a frame
	ldr r2, =0x01000000 // 16.78Mhz CPU cycles a second
	add r0, r0, r1
	cmp r0, r2
	blt 1f

	sub r0, r0, r2
	ldr r1, [r7, #v_time]
	ldr r2, =359999 // 99:59:59
	cmp r1, r2
	beq 1f

	add r1, r1, #1
	str r1, [r7, #v_time]
	ldrb	r1, [r7, #v_uflags]
	mov		r2, #UF_TIME
	orr		r1, r2
	strb	r1, [r7, #v_uflags]
1:
	str		r0, [r7, #v_cycles]
2:

	bl	updateInfo

	ldrb	r0, [r7, #v_uflags]
	cmp	r0, #0
	beq	1f

	bl	textUpdateShadow

1:	mov	r0, #0
	strb	r0, [r7, #v_uflags]

	ldrh	r0, [r7, #v_pattern]
	push {r0}

	bl	mmFrame

	ldrh	r0, [r7, #v_pattern]
	pop {r1}
	cmp r0,r1
	bge 1f

	bl music_continue

1:

	b	main_loop

.pool

/********************************************************************
 * reset( void )
 *
 * Stop music, wait for all keys to be released, then reset visoly card
 ********************************************************************/
						.arm
reset:
	b .no_keys
.reset_stop_music:
	mov	r0, #0
	str	r0, [r7, #v_cycles]
	str	r0, [r7, #v_time]
	mov	r0, #STATE_STOP
	strb	r0, [r7, #v_playstate]
	ldrb	r1, [r7, #v_uflags]
	mov		r2, #UF_TIME
	orr		r1, r2
	strb	r1, [r7, #v_uflags]
	ldr	r0,=mmStop
	bx	r0
.no_keys:
	bl .reset_stop_music
	ldr	r4,=KEYINPUT		// read keys [inversed bits]
	ldr r6,=KEYS_ALL
.nokeys_loop:
	swi	VBlankIntrWaitArm
	bl	mmFrame
	ldrh	r0, [r4]		//
	mvn	r5, r0			//
	tst r5,r6
	bne .nokeys_loop

	// reset offsets
	ldr	r0,=bg_controls_reset
	ldr	r1,=0x4000000+BG0CNT
	mov	r2, #12
	swi	CpuSetArm

.visoly_init_flashcart:
	adr	r0,flash_constants
	ldmia	r0,{r2-r5}
	add	r0,r0,#16

@   WriteRepeat (0x987654, 0x5354, 1);
	ldr	r1,[r0]
	strh	r2,[r1]

@   WriteRepeat ( 0x12345, 0x1234, 500);
	ldr	r1,[r0,#4]
	mov	r12,#500
0:
	strh	r3,[r1]
	subs	r12,r12,#1
	bne	0b

@   WriteRepeat ( 0x12345, 0x5354, 1);
	ldr	r1,[r0,#4]
	strh	r2,[r1]
	
@   WriteRepeat ( 0x12345, 0x5678, 500);
	mov	r12,#500
1:
	strh	r4,[r1]
	subs	r12,r12,#1
	bne	1b

@   WriteRepeat (0x987654, 0x5354, 1);
	ldr	r1,[r0]
	strh	r2,[r1]

@   WriteRepeat ( 0x12345, 0x5354, 1);
	ldr	r1,[r0,#4]
	strh	r2,[r1]

@   WriteRepeat (0x765400, 0x5678, 1);
	ldr	r1,[r0,#12]
	strh	r4,[r1]

@   WriteRepeat ( 0x13450, 0x1234, 1);
	ldr	r1,[r0,#16]
	strh	r3,[r1]

@   WriteRepeat ( 0x12345, 0xabcd, 500);
	ldr	r1,[r0,#4]
	mov	r12,#500
2:
	strh	r5,[r1]
	subs	r12,r12,#1
	bne	2b

@   WriteRepeat 00x987654, 0x5354, 1);
	ldr	r1,[r0]
	strh	r2,[r1]

.visoly_set_rom_start_start:
	mov	r0, #0
	adr	r1,.visoly_set_rom_address
	ldr	r1,[r1]
	strh	r0,[r1]

ez4_reset:
	mov		r3,#0xD200
	mov		r4,#0x1500

	mov		r6,#0x08000000
	add		r7,r6,#0x00020000           ; #0x08020000
	add		r8,r7,#0x00020000			; #0x08040000

	add		r5,r6,#0x02000000
	sub		r1,r5,#0x00800000
	add		r1,r1,#0x00080000			; #0x09880000
	sub		r5,r5,#0x00020000			; #0x09FE0000
	sub		r9,r5,#0x00020000           ; #0x09FC0000
	
	strh	r3,[r5]
	strh	r4,[r6]
	strh	r3,[r7]
	strh	r4,[r8]

	mov		r2,#0x00008000
	strh	r2,[r1]

	strh	r4,[r9]
	
.joint_reset:

	mov	r0, #0 @must be 0 before swi 0x00 is run, otherwise it tries to start from 0x02000000.
	mov	r1, #0x4000000 
	sub	r1,r1,#8  @#0x3FFFFFA
	strh	r0,[r1] @#0x3FFFFFA (mirror of #0x3007FFA)
	mov	r0, #0xfc		@VRAM clear
	swi	0x010000
	swi	0x000000

.visoly_set_rom_address:
	.word	0x096B592E

flash_constants:
	.word	0x5354
	.word	0x1234
	.word	0x5678
	.word	0xabcd

	.word	(0x00987654 * 2) + 0x08000000
	.word	(0x00012345 * 2) + 0x08000000
	.word	(0x00007654 * 2) + 0x08000000
	.word	(0x00765400 * 2) + 0x08000000
	.word	(0x00013450 * 2) + 0x08000000

/********************************************************************
 * keyActions( keys, function )
 *
 * keys = key bitmask
 * function = function to call
 ********************************************************************/
						.thumb_func
keyActions:
	push	{r4,r5,lr}
	mov	r4, r0
	mov	r5, r1

	lsr	r0, r4, #KEY_A			// test A
	bcc	1f				//
	mov	r0, #BUTT_A			//
	bl	.enter_key			//
1:
	lsr	r0, r4, #KEY_B			// test B
	bcc	1f				//
	mov	r0, #BUTT_B			//
	bl	.enter_key			//
1:
	lsr	r0, r4, #KEY_LEFT		// test Left
	bcc	1f				//
	mov	r0, #BUTT_LEFT			//
	bl	.enter_key			//
1:
	lsr	r0, r4, #KEY_RIGHT		// test right
	bcc	1f				//
	mov	r0, #BUTT_RIGHT			//
	bl	.enter_key			//
1:
	lsr	r0, r4, #KEY_START		// test Start
	bcc	1f				//
	ldr r0,=reset
	bx r0
1:
	pop	{r4,r5,pc}

.enter_key:
	bx	r5

.pool
/********************************************************************
 * selPress(butt)
 * 
 * keysdown event
 ********************************************************************/
						.thumb_func
selPress:
	mov	r1, #1
	strb	r1, [r7, #v_press]
	strb	r0, [r7, #v_butt]
	mov	r1, #0				// action[pos]( button, 0 )
						//
jump_keyclick:
	ldrb	r2, [r7, #v_sel]		//
	lsl	r2, #1				//
	add	r2, pc				//
	mov	pc, r2				//
	//nop					//
	b	click_Music			//
	b	click_Pattern			//
	b	click_Tempo			//
	b	click_Pitch			//

/********************************************************************
 * selHold(butt)
 * 
 * keysheld event
 ********************************************************************/
						.thumb_func
selHold:
	ldrb	r1, [r7, #v_press]
	cmp	r1, #0
	beq	1f
	ldrb	r1, [r7, #v_butt]
	cmp	r0, r1
	bne	1f
	mov	r1, #1
	
	b	jump_keyclick
1:	bx	lr

/********************************************************************
 * selRelease(butt)
 * 
 * keysup event
 ********************************************************************/
						.thumb_func
selRelease:

	ldrb	r1, [r7, #v_ptime]
	cmp	r1, #0
	bne	1f
	ldrb	r1, [r7, #v_butt]
	cmp	r0, r1
	bne	1f
	mov	r0, #0
	strb	r0, [r7, #v_press]
1:	bx	lr

/********************************************************************
 * selUp
 *
 * move cursor up
 ********************************************************************/
						.thumb_func
selUp:
	
	ldrb	r0, [r7, #v_sel]		// sel--
	sub	r0, #1				// wrap back to 3
	bpl	.selud_store			//
	mov	r0, #3				//

.selud_store:
	strh	r0, [r7, #v_sel]		//
	mov	r0, #0				// pressed=0
	strb	r0, [r7, #v_press]		//
	bx	lr

/********************************************************************
 * selDown
 *
 * move cursor down
 ********************************************************************/
						.thumb_func
selDown:

	ldrb	r0, [r7, #v_sel]		// sel++
	add	r0, #1				//
	cmp	r0, #3				// wrap back to 0
	ble	.selud_store			//
	mov	r0, #0				//
	b	.selud_store			//

/********************************************************************
 * clickMusicOrPattern(button,held,max,function_a,function_b,flag,value)
 ********************************************************************/
						.thumb_func
clickMusicOrPattern:

	cmp	r2, #0
	bpl	1f
	mov	r0, #0
	strb	r0, [r7, #v_ptime]
	strb	r0, [r7, #v_press]
2:	bx	lr
1:	
	cmp	r1, #1
	beq	2b
	
	mov	r1, #PRESS_TIMEOUT
	strb	r1, [r7, #v_ptime]
	
	cmp	r0, #1
	bgt	.cme_test01
	beq	.cme_b
.cme_a:
	mov	pc, r3
.cme_b:
	mov	pc, r4

.cme_test01:

	ldrh	r1, [r7, r6]

	cmp	r0, #3
	beq	.cme_right
.cme_left:
	sub	r1, #1
	bpl	.cme_set
	mov	r1, r2
	b	.cme_set
.cme_right:
	add	r1, #1
	cmp	r1, r2
	ble	.cme_set
	mov	r1, #0
.cme_set:
	strh	r1, [r7, r6]

	ldrb	r0, [r7, #v_uflags]
	orr	r0, r5
	strb	r0, [r7, #v_uflags]
	
	bx	lr

/********************************************************************
 * click_Music(button,held)
 *
 * click music event
 ********************************************************************/
						.thumb_func
click_Music:
	push	{r4-r6,lr}
	ldrh	r2, [r7, #v_ntracks]
	sub	r2, #1
	adr	r3, music_start
	adr	r4, music_stop
	mov	r5, #UF_MUSIC
	mov	r6, #v_music
	bl	clickMusicOrPattern
	pop	{r4-r6,pc}

						.thumb_func
.align
music_start:
	mov	r0, #0
	str	r0, [r7, #v_cycles]
	str r0, [r7, #v_time]
	ldrb	r1, [r7, #v_uflags]
	mov		r2, #UF_TIME
	orr		r1, r2
	strb	r1, [r7, #v_uflags]
						.thumb_func
music_continue:
	mov	r0, #STATE_PLAY
	strb	r0, [r7, #v_playstate]
	b .pattern_total
.start_music:
	ldrh	r0, [r7, #v_music]
	mov	r1, #0		// <- loop
	ldr	r3,=mmStart
	bx	r3

.pattern_total:
	push {lr}
	bl .start_music
	// Find pattern total
	mov r1, #0
	mov r2, #0
	ldr	r0, =mmLayerMain
	ldr r0, [r0, #MPL_SONGADR]
	add r0, #C_MAS_ORDER
0:
	cmp r1, #255
	beq 1f
	ldrb r3, [r0, r2]
	add r2, #1
	cmp r3, #254
	beq 2f
	add r1, #1
2:
	cmp r3, #255
	bne 0b
1:
	sub r1, #2
	strh r1, [r7, #v_npatterns]
	pop {r3}
	bx r3

.align
music_pauseplay:
	ldrb	r0, [r7, #v_playstate]
	cmp r0, #STATE_STOP
	beq .play
	cmp r0, #STATE_PLAY
	beq .pause
.resume:
	mov r0, #STATE_PLAY
	strb	r0, [r7, #v_playstate]
	ldr r0, =mmResume
	bx r0
.pause:
	mov	r0, #STATE_PAUSE
	strb	r0, [r7, #v_playstate]
	ldr	r0, =mmPause
	bx r0
.play:
	b music_start

.align
music_stop:
	mov	r0, #0
	str	r0, [r7, #v_cycles]
	str r0, [r7, #v_time]
	ldrb	r1, [r7, #v_uflags]
	mov		r2, #UF_TIME
	orr		r1, r2
	strb	r1, [r7, #v_uflags]
	mov	r0, #STATE_STOP
	strb	r0, [r7, #v_playstate]
	ldr	r0,=mmStop
	bx	r0

/********************************************************************
 * click_Pattern(button,held)
 *
 * click pattern event
 ********************************************************************/
						.thumb_func
click_Pattern:
	push	{r4-r6,lr}

	ldrb	r2, [r7, #v_playstate]
	ldrh	r3, [r7, #v_pattern]
	push {r2,r3}

	ldrh	r2, [r7, #v_npatterns]
	ldr	r3, =music_pauseplay
	ldr	r4, =music_stop
	mov	r5, #UF_PATTERN
	mov	r6, #v_pattern
	bl	clickMusicOrPattern

	pop {r2,r3}
	cmp r2, #STATE_PLAY
	bne 1f
	ldrb	r2, [r7, #v_playstate]
	cmp r2, #STATE_PLAY
	bne 1f
	ldrh r0, [r7, #v_pattern]
	b 2f
1:
	strh r3, [r7, #v_pattern]
	b .end
.position:
	ldr r1, =mmPosition
	bx r1
2:
	push {r0}
	bl music_continue //called to reset channels
	pop {r0}
	bl .position
.end:
	pop	{r4-r6,pc}


/***************************************************************
 * click tempo/pitch
 ***************************************************************/
.macro CLICK_VALUE value, flag, function

	push	{lr}

	cmp	r0, #BUTT_A			// ignore A
	bne	1f				//
	strb	r0, [r7, #v_press]		//
	strb	r0, [r7, #v_ptime]		//
3:	pop	{pc}
1:	
	ldrh	r2, [r7, #\value]
	cmp	r0, #BUTT_LEFT			// test B,left,right

	beq	1f				//
	bgt	2f				//
//------------------------------------------------
// B pressed
//------------------------------------------------
	cmp	r1, #0				// reset value
	bne	3b				//
	mov	r0, #PRESS_TIMEOUT
	strb	r0, [r7, #v_ptime]

	mov	r2, #1				//
	lsl	r2, #10				//
	b	4f				//

//------------------------------------------------
1: // Left pressed
//------------------------------------------------
	sub	r2, #SCROLL_RATE		// scroll down
	lsr	r0, r2, #9			//
	bne	3f				// clip to 512
	mov	r2, #1				//
	lsl	r2, #9				//
	b	3f				//

//------------------------------------------------
2: // Right pressed
//------------------------------------------------
	add	r2, #SCROLL_RATE		// scroll up
	lsr	r0, r2, #11			//
	beq	3f				// clip to 2048
	mov	r2, #1				//
	lsl	r2, #11				//

//------------------------------------------------
3: // store value
//------------------------------------------------

	mov	r0, #0
	strb	r0, [r7, #v_ptime]

4:
	strh	r2, [r7, #\value]
	
	ldrb	r0, [r7, #v_uflags]
	mov	r1, #\flag
	orr	r0, r1
	strb	r0, [r7, #v_uflags]

	mov	r0, r2
	bl	\function

	POP	{pc}
.endm

/********************************************************************
 * click_Tempo(button,held)
 *
 * click tempo event
 ********************************************************************/
						.thumb_func
click_Tempo:
	
	CLICK_VALUE v_tempo, UF_TEMPO, mmSetModuleTempo

/********************************************************************
 * click_Pitch(button,held)
 *
 * click pitch event
 ********************************************************************/
						.thumb_func
click_Pitch:
	
	CLICK_VALUE v_pitch, UF_PITCH, mmSetModulePitch

/********************************************************************
 * textString( x, y, string )
 *
 * render string
 ********************************************************************/
textString:
	
	lsl	r1, #5
	add	r0, r1
	lsl	r0, #1
	ldr	r1,=BG0MAP
	add	r1, r0
	b	2f

1:	strh	r0, [r1]
	add	r1, #2
2:	ldrb	r0, [r2]
	add	r2, #1
	cmp	r0, #0
	bne	1b

.end_string:
	bx	lr

/********************************************************************
 * textUpdateShadow()
 *
 * Update text shadow layer
 ********************************************************************/
textUpdateShadow:

	push	{r4}

	mov	r4, #1
	lsl	r4, #9
	
	ldr	r0,=BG0MAP
	ldr	r1,=BG1MAP

	ldr	r2,=0x10001000
	
1:	ldmia	r0!, {r3}
	add	r3, r2
	stmia	r1!, {r3}
	
	sub	r4, #1
	bne	1b

	pop	{r4}
	bx	lr

/********************************************************************
 * updateFinge
 *
 * Update finge sprite
 ********************************************************************/
						.thumb_func
updateFinge:

	ldrb	r0, [r7, #v_ptime]		// if timeout
	sub	r0, #1				//   timeout--
	bmi	1f				//   pressed=timeout
	strb	r0, [r7, #v_ptime]		//  
	cmp	r0, #0				//
	bne	1f				//
	strb	r0, [r7, #v_press]		//
1:

	ldr	r0,=0x7000000			// y = 56+sel*16
	mov	r1, #56				//
	ldrb	r2, [r7, #v_sel]		//
	lsl	r2, #4				//
	add	r1, r2				//
	strh	r1, [r0]			//

	ldr	r2,=(1<<10)+1			// shadow, +1 y, blend.
	add	r1, r2				//
	strh	r1, [r0, #8]			//
	
	mov	r1, #44				// x = 44 + pressed
	ldrb	r2, [r7, #v_press]		//
	cmp	r2, #0				//
	beq	1f				//
	add	r1, #1				//
1:	strh	r1, [r0, #2]			//
	
	add	r1, #1				// shadow
	strh	r1, [r0, #10]			//
	
	mov	r1, #0				// tile:0	
	strh	r1, [r0, #4]			//

	mov	r1, #1				// shadow
	lsl	r1, #12				//
	strh	r1, [r0, #12]			//

	bx	lr

/****************************************************************************
 * drawNumber(number,target,filler)
 * drawNumberP(number,target)
 *
 * draw 2 to 3 digit number
 * P: convert 0->2048 into 0->100 (percent) 
 ****************************************************************************/
						.thumb_func
drawNumberP:
	mov	r2, #100
	mul	r0, r2
	add	r0, #255
	add	r0, #255
	lsr	r0, #10
	mov r2, #0
						.thumb_func
drawNumber:
	push {r2}
	mov	r2, r1
	mov	r1, #10
	swi	Div
	add	r1, #'0'
	strh	r1, [r2, #4]
	
	cmp	r0, #10
	blt	1f
	mov	r1, #10
	swi	Div
	add	r1, #'0'
	strh	r1, [r2, #2]
	add	r0, #'0'
	strh	r0, [r2, #0]
	pop {r0}
	b	.dnexit
1:
	add	r0, #'0'
	strh	r0, [r2, #2]
	pop {r0}
	strh	r0, [r2, #0]

.dnexit:
	bx	lr

/****************************************************************************
 * updateInfo()
 *
 * update onscreen information
 ****************************************************************************/
						.thumb_func
updateInfo:
	push	{r4-r7,lr}
	ldrb	r4, [r7, #v_uflags]

	lsr	r4, #1				// test&update Music
	bcc	1f				//
						//
	ldrh	r0, [r7, #v_ntracks]		//
	cmp	r0, #0				//
	beq	1f				//
						//
	ldrh	r0, [r7, #v_music]		//
	ldr	r1,=BG0MAP + 20*2 + 7*32*2	//
	mov r2, #0
	bl	drawNumber			//
						//
1:						//

	lsr	r4, #1				// test&update Pattern
	bcc	1f				//
						//
	ldrh	r0, [r7, #v_pattern] //current_pattern
	ldr	r1,=BG0MAP + 16*2 + 9*32*2	//
	mov r2, #0
	bl	drawNumber			//
	ldrh	r0, [r7, #v_npatterns] //total_patterns
	ldr	r1,=BG0MAP + 20*2 + 9*32*2	//
	mov r2, #0
	bl	drawNumber			//
						//
1:						//

	lsr	r4, #1				// test&update Tempo
	bcc	1f				//
						//
	ldrh	r0, [r7, #v_tempo]		// convert tempo into percent
	ldr	r1,=BG0MAP + 19*2 + 11*32*2	//
	bl	drawNumberP			//
						//
1:						//

	lsr	r4, #1				// test&update Pitch
	bcc	1f				//
						//
	ldrh	r0, [r7, #v_pitch]		//
	ldr	r1,=BG0MAP + 19*2 + 13*32*2	//
	bl	drawNumberP			//
						//
1:						//
	lsr	r4, #1				// test&update Time
	bcc	1f				//

	ldr r0, [r7, #v_time]
	mov r1, #60
	swi Div

	mov r5, r0
	mov r0, r1
	ldr r1, =BG0MAP + 16*2 + 4*32*2
	mov r2, #':'
	bl drawNumber // seconds

	mov r0, r5
	mov r1, #60
	swi Div

	mov r5, r0
	mov r0, r1
	ldr r1, =BG0MAP + 13*2 + 4*32*2
	mov r2, #':'
	bl drawNumber	// minutes

	mov r0, r5
	ldr r1, =BG0MAP + 10*2 + 4*32*2
	mov r2, #0
	bl drawNumber	// hours
1:

	pop	{r4-r7,pc}

/****************************************************************************
 *
 * Interrupts
 *
 ****************************************************************************/

	.ALIGN
	.ARM

//---------------------------------------------------------------------------
intr_main:
//---------------------------------------------------------------------------

	mov	r0, #0x4000000		@ REGISTERS
	ldr	r1, [r0,#0x200]		@ IE IF
	and	r1, r1, r1, lsr#16	@ and together (occured&enabled)
	add	r2, r0, #0x200		@ acknowledge interrupts 
	strh	r1, [r2, #2]		@ (store to IF)

	ldrh	r2, [r0, #-8]		@ ORR to 0x3007FF8
	orr	r2, r2, r1		@ required for SWI 5
	strh	r2, [r0, #-8]		@ ..

        @and     r1, r3, r3, lsr #16	@ r1 = IE & IF

	tst	r1, #0x80		@ comm interrupt?
	bne	burst_entry

	ldr	r0,=DMA0SAD		@ reset bg DMA
	mov	r1, #0			@
	str	r1, [r0, #0x8]		@
	ldr	r1,=bg_color_data	@
	ldr	r2,=0x5000000		@
	stmia	r0!, {r1,r2}		@
	ldr	r1,=0xA2400001		@
	str	r1, [r0]		@

	ldr	r1,=31<<10		@ initialize first color
	strh	r1, [r2]		@
	
1:	ldr	r0,=mmVBlank		@ otherwise this interrupt is a vblank interrupt
	bx	r0			@ pass control to mmVBlank


/****************************************************************************
 * 
 * Burstboot Backdoor
 *
 ****************************************************************************/

burst_entry:
	ldr	r2,[r0,#0x120]
	ldr	r3, msg_brst	@ check for signal
	cmp	r2,r3
	beq	burst_boot
	bx	lr

@-------------------------------------------------------------------------------------
burst_boot:
@-------------------------------------------------------------------------------------

	mov	r2, #0
	str	r2, [r0, #0xC6]	@ disable hdma
	str	r2, [r0, #0xC4]	@ disable dma1
	str	r2, [r0, #0xD0]	@ disable dma2
	strh	r2, [r0, #0]	@ disable screen
	str	r2, [r0, #0x84]	@ disable sound
	
	@ receive loader from pc
	ldr	r4, msg_okay
	bl	sio_transfer	@ send okay signal, receive transfer length 
	mov	r2, r1		@
	ldr	r3,=__bss_start	@ transfer program to iwram
	mov	r4, #0		@ crc counter...
.bb_transfer:
	bl	sio_transfer	@ transfer crc, receive word
	str	r1, [r3], #4	@ store word
	add	r4, r4, r1	@ add to crc
	subs	r2, r2, #4	@ count
	bhi	.bb_transfer	@ loop
	bl	sio_transfer	@ send CRC value to master..
	ldr	r3,=__bss_start	@ 
	bx	r3		@ start transfer program
	
@--------------------------------------------------------------------------------------
sio_transfer:
@--------------------------------------------------------------------------------------

	@ r4 = data
	@ r0,=0x4000000
	@ returns r1=received data
	str	r4, [r0, #0x120]
	ldr	r1, [r0, #0x128]
	orr	r1, r1, #0x80
	str	r1, [r0, #0x128]
.st_wait:
	ldr	r1, [r0, #0x128]
	tst	r1, #0x80
	bne	.st_wait
	ldr	r1, [r0, #0x120]
	bx	lr

@------------------------------------
@ tables
@------------------------------------

.align
msg_boot:
	.byte	'B','O','O','T'

msg_okay:
	.byte	'O','K','A','Y'

msg_brst:
	.byte	'B','R','S','T'

@-----------------------------------
.align

sndsystem:
	.word	5		@ 21khz mixing
	.word	32		@ 32ch
	.word	32
	.word	mod_ch
	.word	act_ch
	.word	mix_ch
	.word	mixing_buffer
	.word	wave_buffer
	.word	0		@ soundbank, change during runtime
.align
unpack_tiles:
	.hword	768		// size
	.byte	1		// src width
	.byte	4		// dest width
	.word	0		// color


.align
bg_controls:
	.hword	0x0300		// cnt0
	.hword	0x0400		// cnt1	
	.hword	0x0500		// cnt2
	.hword	0x0600		// cnt3
	.hword	0			// hofs0
	.hword	0			// vofs0
	.hword	-1		// hofs1
	.hword	-1		// vofs1
	.hword	-32		// hofs2
	.hword	-20		// vofs2
	.hword	-38		// hofs3
	.hword	-26		// vofs3

.align
bg_controls_reset:
	.hword	0x0000		// cnt0
	.hword	0x0000		// cnt1	
	.hword	0x0000		// cnt2
	.hword	0x0000		// cnt3
	.hword	0		// hofs0
	.hword	0		// vofs0
	.hword	0		// hofs1
	.hword	0		// vofs1
	.hword	0		// hofs2
	.hword	0		// vofs2
	.hword	0		// hofs3
	.hword	0		// vofs3

.end
